/*
 * Copyright (C) 1999 Colin Van Dyke
 * Copyright (C) 1999-2002 Vojtech Pavlik
 * Copyright (C) 2006 Jean-Yves Lefort <jylefort@FreeBSD.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/cdefs.h>

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/uio.h>
#include <sys/poll.h>
#include <sys/module.h>
#include <sys/types.h>
#include <sys/malloc.h>
#include <sys/fcntl.h>
#include <sys/queue.h>
#include <sys/time.h>
#include <sys/joystick.h>
#include <machine/bus.h>
#include <dev/joy/joyvar.h>

#include "joystick.h"

#define N_ELEMENTS(array)		(sizeof(array) / sizeof((array)[0]))

#define LINUX_JS_MAX_BUTTONS		LINUX_JS_MAX_AXES
#define LINUX_JS_MAX_DEVICES		(LINUX_JS_MAX_AXES / 2)

#define LINUX_JS_INVALID_AXIS		0x80000000 /* sys/dev/joy/joy.c */

#define LINUX_JS_MAP_AXIS(num)		(linux_js_ax_map[(num)])
#define LINUX_JS_MAP_BTN(num)		(linux_js_btn_map[(num)] - BTN_MISC)

struct linux_js_queue_entry
{
  struct js_event			event;
  STAILQ_ENTRY(linux_js_queue_entry)	next;
};

STAILQ_HEAD(linux_js_queue, linux_js_queue_entry);

struct linux_js_softc 
{
  struct cdev		*devices[LINUX_JS_MAX_DEVICES];
  int			n_devices;
  int			axes[LINUX_JS_MAX_AXES];
  int			buttons[LINUX_JS_MAX_BUTTONS];
  struct linux_js_queue	queue;
  struct timeval	open_time;
};

static struct js_corr	linux_js_corr[LINUX_JS_MAX_AXES];
static uint8_t		linux_js_ax_map[ABS_MAX + 1];
static uint16_t		linux_js_btn_map[KEY_MAX - BTN_MISC + 1];
static struct cdev	*linux_js_dev;

static int linux_js_read_joystick_data (struct thread *td,
					struct cdev *joy_dev,
					struct joystick *data);

static int linux_js_correct (int value, struct js_corr *corr);

static void linux_js_handle_axis (struct linux_js_softc *js,
				  uint32_t *timestamp,
				  int num,
				  int value);
static void linux_js_handle_button (struct linux_js_softc *js,
				    uint32_t *timestamp,
				    int num,
				    int value);

static void linux_js_queue_init (struct linux_js_softc *js, struct thread *td);
static void linux_js_queue_fill (struct linux_js_softc *js, struct thread *td);
static void linux_js_queue_append (struct linux_js_softc *js,
				   uint32_t *timestamp,
				   int16_t value,
				   uint8_t type,
				   uint8_t number);

static int
linux_js_open (struct cdev *dev, int flags, int fmt, struct thread *td)
{
  struct linux_js_softc *js;
  devclass_t joy_class;
  int i;

  if (dev->si_drv1)
    return EBUSY;

  joy_class = devclass_find("joy");
  if (! joy_class)
    return ENXIO;

  js = malloc(sizeof(struct linux_js_softc), M_DEVBUF, M_WAITOK | M_ZERO);
  /* M_WAITOK -> malloc cannot return NULL */

  for (i = 0; i < LINUX_JS_MAX_DEVICES; i++)
    {
      struct joy_softc *softc;

      softc = devclass_get_softc(joy_class, i);
      if (softc && softc->d->si_devsw->d_open(softc->d, O_RDONLY, 0, td) == 0)
	{
	  js->devices[js->n_devices] = softc->d;
	  js->n_devices++;
	}
    }

  if (js->n_devices > 0)
    {
      dev->si_drv1 = js;

      microtime(&js->open_time);
      linux_js_queue_init(js, td);

      return 0;
    }
  else
    {
      free(js, M_DEVBUF);

      return ENXIO;
    }
}

static int
linux_js_close (struct cdev *dev, int flags, int fmt, struct thread *td)
{
  struct linux_js_softc *js = dev->si_drv1;
  struct linux_js_queue_entry *entry;
  int i;

  for (i = 0; i < js->n_devices; i++)
    js->devices[i]->si_devsw->d_close(js->devices[i], 0, 0, td);

  while ((entry = STAILQ_FIRST(&js->queue)))
    {
      STAILQ_REMOVE_HEAD(&js->queue, next);
      free(entry, M_DEVBUF);
    }

  free(js, M_DEVBUF);
  dev->si_drv1 = NULL;

  return 0;
}

static int
linux_js_read (struct cdev *dev, struct uio *uio, int flag)
{
  struct linux_js_softc *js = dev->si_drv1;
  struct linux_js_queue_entry *entry;
  int status;

  entry = STAILQ_FIRST(&js->queue);
  if (! entry)
    {
      linux_js_queue_fill(js, uio->uio_td);
      entry = STAILQ_FIRST(&js->queue);

      if (! entry)
	{
	  if ((flag & O_NONBLOCK) != 0)
	    return EWOULDBLOCK;

	  do
	    {
	      static int dummy;

	      if (tsleep(&dummy, PRIBIO | PCATCH, "joyslp", hz / 50) != EWOULDBLOCK)
		return EINTR;

	      linux_js_queue_fill(js, uio->uio_td);
	      entry = STAILQ_FIRST(&js->queue);
	    }
	  while (! entry);
	}
    }

  status = uiomove((caddr_t) &entry->event, sizeof(struct js_event), uio);

  STAILQ_REMOVE_HEAD(&js->queue, next);
  free(entry, M_DEVBUF);

  return status;
}

static int
linux_js_ioctl (struct cdev *dev, u_long cmd, caddr_t data, int flag, struct thread *td)
{
  struct linux_js_softc *js = dev->si_drv1;

  switch (cmd)
    {
    case JSIOCGVERSION:
      *(uint32_t *) data = JS_VERSION;
      break;

    case JSIOCGAXES:
    case JSIOCGBUTTONS:
      *(uint8_t *) data = js->n_devices * 2;
      break;

    case JSIOCSCORR:
      memcpy(linux_js_corr, data, sizeof(linux_js_corr));
      break;

    case JSIOCGCORR:
      memcpy(data, linux_js_corr, sizeof(linux_js_corr));
      break;

    case JSIOCSAXMAP:
      {
	int n_axes = js->n_devices * 2;
	uint8_t *ax_map = (uint8_t *) data;
	int i;

	for (i = 0; i < n_axes; i++)
	  {
	    if (ax_map[i] > ABS_MAX)
	      return EINVAL;
	    linux_js_ax_map[i] = ax_map[i];
	  }
      }
      break;

    case JSIOCGAXMAP:
      memcpy(data, linux_js_ax_map, sizeof(linux_js_ax_map));
      break;

    case JSIOCSBTNMAP:
      {
	int n_buttons = js->n_devices * 2;
	uint16_t *btn_map = (uint16_t *) data;
	int i;

	for (i = 0; i < n_buttons; i++)
	  {
	    if (btn_map[i] < BTN_MISC || btn_map[i] > KEY_MAX)
	      return EINVAL;
	    linux_js_btn_map[i] = btn_map[i];
	  }
      }
      break;

    case JSIOCGBTNMAP:
      memcpy(data, linux_js_btn_map, sizeof(linux_js_btn_map));
      break;

    default:
      if (IOCBASECMD(cmd) == JSIOCGNAME(0))
	strlcpy(data, "FreeBSD analog joystick(s)", IOCPARM_LEN(cmd));
      else
	return ENOTTY;
    }

  return 0;
}

static int
linux_js_poll (struct cdev *dev, int events, struct thread *td)
{
  struct linux_js_softc *js = dev->si_drv1;
  int revents = 0;

  if ((events & (POLLIN | POLLRDNORM)) != 0)
    {
      if (STAILQ_EMPTY(&js->queue))
	linux_js_queue_fill(js, td);

      if (! STAILQ_EMPTY(&js->queue))
	revents |= events & (POLLIN | POLLRDNORM);
    }

  return revents;
}

static int
linux_js_read_joystick_data (struct thread *td,
			     struct cdev *joy_dev,
			     struct joystick *data)
{
  struct uio auio;
  struct iovec aiov;

  aiov.iov_base = data;
  aiov.iov_len = sizeof(struct joystick);
  auio.uio_iov = &aiov;
  auio.uio_iovcnt = 1;
  auio.uio_offset = 0;
  auio.uio_resid = sizeof(struct joystick);
  auio.uio_segflg = UIO_SYSSPACE;
  auio.uio_rw = UIO_READ;
  auio.uio_td = td;

  if (joy_dev->si_devsw->d_read(joy_dev, &auio, 0) == 0)
    {
      if (data->x == LINUX_JS_INVALID_AXIS)
	data->x = 0;
      if (data->y == LINUX_JS_INVALID_AXIS)
	data->y = 0;

      return 1;
    }
  else
    return 0;
}

static int
linux_js_correct (int value, struct js_corr *corr)
{
  switch (corr->type)
    {
    case JS_CORR_NONE:
      break;

    case JS_CORR_BROKEN:
      value = value > corr->coef[0]
	? (value < corr->coef[1]
	   ? 0
	   : ((corr->coef[3] * (value - corr->coef[1])) >> 14))
	: ((corr->coef[2] * (value - corr->coef[0])) >> 14);
      break;
    }

  if (value < -32767)
    return -32767;
  else if (value > 32767)
    return 32767;
  else
    return value;
}

static void
linux_js_handle_axis (struct linux_js_softc *js,
		      uint32_t *timestamp,
		      int num,
		      int value)
{
  if (abs(value - js->axes[num]) > linux_js_corr[num].prec)
    {
      js->axes[num] = value;
      linux_js_queue_append(js, timestamp, linux_js_correct(value, &linux_js_corr[num]), JS_EVENT_AXIS, num);
    }
}

static void
linux_js_handle_button (struct linux_js_softc *js,
			uint32_t *timestamp,
			int num,
			int value)
{
  if (value != js->buttons[num])
    {
      js->buttons[num] = value;
      linux_js_queue_append(js, timestamp, value, JS_EVENT_BUTTON, num);
    }
}

static void
linux_js_queue_init (struct linux_js_softc *js, struct thread *td)
{
  int i;
  uint32_t timestamp = 0;

  STAILQ_INIT(&js->queue);

  for (i = 0; i < js->n_devices; i++)
    {
      struct joystick data;
      int num1 = i * 2;
      int num2 = i * 2 + 1;

      if (! linux_js_read_joystick_data(td, js->devices[i], &data))
	memset(&data, 0, sizeof(data));

      linux_js_queue_append(js, &timestamp, linux_js_correct(data.x, &linux_js_corr[LINUX_JS_MAP_AXIS(num1)]), JS_EVENT_AXIS | JS_EVENT_INIT, LINUX_JS_MAP_AXIS(num1));
      linux_js_queue_append(js, &timestamp, linux_js_correct(data.y, &linux_js_corr[LINUX_JS_MAP_AXIS(num2)]), JS_EVENT_AXIS | JS_EVENT_INIT, LINUX_JS_MAP_AXIS(num2));
      linux_js_queue_append(js, &timestamp, data.b1, JS_EVENT_BUTTON | JS_EVENT_INIT, LINUX_JS_MAP_BTN(num1));
      linux_js_queue_append(js, &timestamp, data.b2, JS_EVENT_BUTTON | JS_EVENT_INIT, LINUX_JS_MAP_BTN(num2));
    }
}

static void
linux_js_queue_fill (struct linux_js_softc *js, struct thread *td)
{
  int i;
  uint32_t timestamp = 0;

  for (i = 0; i < js->n_devices; i++)
    {
      struct joystick data;

      if (linux_js_read_joystick_data(td, js->devices[i], &data))
	{
	  int num1 = i * 2;
	  int num2 = i * 2 + 1;

	  linux_js_handle_axis(js, &timestamp, LINUX_JS_MAP_AXIS(num1), data.x);
	  linux_js_handle_axis(js, &timestamp, LINUX_JS_MAP_AXIS(num2), data.y);
	  linux_js_handle_button(js, &timestamp, LINUX_JS_MAP_BTN(num1), data.b1);
	  linux_js_handle_button(js, &timestamp, LINUX_JS_MAP_BTN(num2), data.b2);
	}
    }
}

static void
linux_js_queue_append (struct linux_js_softc *js,
		       uint32_t *timestamp,
		       int16_t value,
		       uint8_t type,
		       uint8_t number)
{
  struct linux_js_queue_entry *entry;

  if (*timestamp == 0)
    {
      struct timeval tval;

      microtime(&tval);
      timevalsub(&tval, &js->open_time);

      *timestamp = tval.tv_sec * 1000 + tval.tv_usec / 1000;
    }

  entry = malloc(sizeof(struct linux_js_queue_entry), M_DEVBUF, M_NOWAIT);
  if (entry)
    {
      entry->event.time = *timestamp;
      entry->event.value = value;
      entry->event.type = type;
      entry->event.number = number;

      STAILQ_INSERT_TAIL(&js->queue, entry, next);
    }
}

static struct cdevsw linux_js_cdevsw = {
  .d_version =	D_VERSION,
  .d_flags =	D_NEEDGIANT,
  .d_name =	"js",
  .d_open =	linux_js_open,
  .d_close =	linux_js_close,
  .d_read =	linux_js_read,
  .d_ioctl =	linux_js_ioctl,
  .d_poll =	linux_js_poll
};

static int
linux_js_load (module_t mod, int cmd, void *arg)
{
  switch (cmd)
    {
    case MOD_LOAD:
      {
	int i;

	bzero(linux_js_corr, sizeof(linux_js_corr));

	for (i = 0; i < N_ELEMENTS(linux_js_ax_map); i++)
	  linux_js_ax_map[i] = i;
	for (i = 0; i < N_ELEMENTS(linux_js_btn_map); i++)
	  linux_js_btn_map[i] = BTN_MISC + i;

	linux_js_dev = make_dev(&linux_js_cdevsw, 0, UID_ROOT, GID_WHEEL, 0644, "input/js%d", 0);
	if (! linux_js_dev)
	  return ENOMEM;
      }
      break;

    case MOD_UNLOAD:
    case MOD_SHUTDOWN:
      if (linux_js_dev->si_drv1)
	return EBUSY;
      destroy_dev(linux_js_dev);
      break;

    default:
      return EOPNOTSUPP;
    }

  return 0;
}

DEV_MODULE(linux_js, linux_js_load, NULL);
